///--------------------------------------------------------------------------------------
//	SpriteWorld.h
//
//	Portions are copyright:  1991-94 Tony Myles, All rights reserved worldwide.
//
//	Description:	constants, structures, and prototypes for sprite worlds
///--------------------------------------------------------------------------------------


#ifndef __SPRITEWORLD__
#define __SPRITEWORLD__

#ifndef __SWCOMMON__
#include "SWCommonHeaders.h"
#endif

#ifndef __SPRITELAYER__
#include "SpriteLayer.h"
#endif

#ifndef __SPRITEFRAME__
#include "SpriteFrame.h"
#endif

#ifndef __SPRITE__
#include "Sprite.h"
#endif

#ifndef __TILING__
#include "Tiling.h"
#endif

#ifndef VBLUPP
#define VBLUPP ProcPtr
#endif

///--------------------------------------------------------------------------------------
//	sprite world error constants
///--------------------------------------------------------------------------------------

enum
{
	kSystemTooOldErr = 100,		// < System 7.0
	kMaxFramesErr,				// attempt to exceed maximum number of frames for a sprite
	kInvalidFramesIndexErr,		// frame index out of range
	kNotCWindowErr,				// attempt to make a SpriteWorld from non-color WindowPtr
	kNilParameterErr,			// nil SpritePtr, FramePtr, etc.
	kWrongDepthErr,				// invalid pixel size for attempted function
	kWrongMaskErr,				// invalid mask type for attempted function
	kOutOfRangeErr,				// tileID or tileMap value out of bounds
	kTilingNotInitialized,		// tiling hasn't been initialized
	kTilingAlreadyInitialized,	// tiling already initialized; can't be initialized again
	kNullTileMapErr				// no TileMap has ever been created/loaded
};


///--------------------------------------------------------------------------------------
//	sprite world type definitions
///--------------------------------------------------------------------------------------

typedef SW_FUNC void (*WorldMoveProcPtr)(
	SpriteWorldPtr spriteWorldP,
	SpritePtr followSpriteP);

///--------------------------------------------------------------------------------------
//	sprite world data structure
///--------------------------------------------------------------------------------------

struct SpriteWorldRec
{
	SpriteLayerPtr headSpriteLayerP;	// head of the sprite layer linked list
	SpriteLayerPtr tailSpriteLayerP;	// tail of the sprite layer linked list
	
	FramePtr backFrameP;				// frame for the background
	FramePtr workFrameP;				// work, or "mixing" frame
	FramePtr windowFrameP;				// frame for drawing to the screen
	
	DrawProcPtr offscreenDrawProc;		// callback for erasing sprites offscreen
	DrawProcPtr screenDrawProc;			// callback for drawing sprite pieces onscreen
	
	Rect	windRect;					// holds windowFrameP->frameRect for easier access
	Rect	backRect;					// holds backFrameP->frameRect for easier access
	
	Rect	visScrollRect;				// rect that is copied to screen when scrolling
	Rect	oldVisScrollRect;			// visScrollRect from last frame
	Rect	offscreenScrollRect;		// same as visScrollRect, but local to offscreen
	short	horizScrollRectOffset;		// offset from offscreenScrollRect to visScrollRect
	short	vertScrollRectOffset;		// offset from offscreenScrollRect to visScrollRect
	short	horizScrollDelta;			// horizontal scrolling delta
	short	vertScrollDelta;			// vertical scrolling delta
	Rect	scrollRectMoveBounds;		// move bounds for visScrollRect
	
	WorldMoveProcPtr worldMoveProc;		// pointer to the scrolling world move procedure
	SpritePtr followSpriteP;			// pointer to the "follow sprite", or NULL
	
	TileMapPtr	tileMap;				// two-dimensional tile map array
	Boolean		tilingIsInitialized;	// has the tiling been initialized yet?
	short		numTileMapRows;			// number of rows in tileMap array
	short		numTileMapCols;			// number of cols in tileMap array
	short		*offscreenTileMap;		// offscreen tile map array
	short		numOffscreenTileMapRows;// number of rows in offscreenTileMap array
	short		numOffscreenTileMapCols;// number of cols in offscreenTileMap array
	FramePtr	*tileFrameArray;		// array of tile framePtrs
	short		*curTileImage;			// array specifying the current frame of each tile
	short		maxNumTiles;			// number of elements in tileFrameArray
	short		tileWidth;				// width of each tile
	short		tileHeight;				// height of each tile
	Boolean		tilingIsOn;				// are the tiling routines turned on?
	short		numTilesChanged;		// number of rects in updateRect array to update
	Rect		*changedTiles;			// array of rects of tiles that changed
	TileChangeProcPtr tileChangeProc;	// pointer to tile frame changing procedure
	DrawProcPtr tileMaskDrawProc;		// drawProc for drawing masked tiles above sprites
	
	GDHandle 	mainSWGDH;				// GDH of SpriteWorld's window
	short 		pixelDepth;				// SpriteWorld's depth 
	
	unsigned long	runningTimeCount;	// running total time in milliseconds
	UnsignedWide 	lastMicroseconds;	// value of previous Microseconds() call
	Boolean 		frameHasOccured;	// Has the SpriteWorld been processed?
	short 			fpsTimeInterval;	// milliseconds per frame of animation (1000/fps)
	unsigned long 	timeOfLastFrame;	// time (from runningTimeCount) of last frame
	VBLTaskRec 		vblTaskRec;			// extended VBLTask record
	Boolean 		usingVBL;			// is the VBL task installed?

	Boolean anySpritesNeedRemoving;		// flag that at least one Sprite needs to be removed
		
	long userData;						// reserved for user
};


///--------------------------------------------------------------------------------------
//	sprite world function prototypes
///--------------------------------------------------------------------------------------


#ifdef __cplusplus
extern "C" {
#endif

SW_FUNC OSErr SWEnterSpriteWorld(void);

SW_FUNC void SWExitSpriteWorld(void);

SW_FUNC OSErr SWCreateSpriteWorld(
	GDHandle mainGDH,
	SpriteWorldPtr *spriteWorldP,
	FramePtr windowFrameP,
	FramePtr backFrameP,
	FramePtr workFrameP);

SW_FUNC OSErr SWCreateSpriteWorldFromWindow(
	SpriteWorldPtr*	spriteWorldP,
	CWindowPtr		srcWindowP,
	Rect*			worldRectP,
	Rect*			offscreenRectP);

SW_FUNC void SWDisposeSpriteWorld(
	SpriteWorldPtr spriteWorldP);

SW_FUNC void SWAddSpriteLayer(
	SpriteWorldPtr spriteWorldP,
	SpriteLayerPtr spriteLayerP);

SW_FUNC void SWRemoveSpriteLayer(
	SpriteWorldPtr spriteWorldP,
	SpriteLayerPtr spriteLayerP);

SW_FUNC void SWSwapSpriteLayer(
	SpriteWorldPtr spriteWorldP,
	SpriteLayerPtr srcSpriteLayerP,
	SpriteLayerPtr dstSpriteLayerP);

SW_FUNC SpriteLayerPtr SWGetNextSpriteLayer(
	SpriteWorldPtr spriteWorldP,
	SpriteLayerPtr curSpriteLayerP);

SW_FUNC void SWLockSpriteWorld(
	SpriteWorldPtr spriteWorldP);

SW_FUNC void SWUnlockSpriteWorld(
	SpriteWorldPtr spriteWorldP);

SW_FUNC void SWSetPortToBackground(
	SpriteWorldPtr spriteWorldP);

SW_FUNC void SWSetPortToWorkArea(
	SpriteWorldPtr	spriteWorldP);

SW_FUNC void SWSetPortToWindow(
	SpriteWorldPtr spriteWorldP);

SW_FUNC OSErr SWSetSpriteWorldOffscreenDrawProc(
	SpriteWorldPtr spriteWorldP,
	DrawProcPtr offscreenProc);

SW_FUNC OSErr SWSetSpriteWorldScreenDrawProc(
	SpriteWorldPtr spriteWorldP,
	DrawProcPtr drawProc);
	
SW_FUNC void SWStdWorldDrawProc(
	FramePtr srcFrameP,
	FramePtr dstFrameP,
	Rect* srcRect,
	Rect* dstRect);

SW_FUNC void SWCopyBackgroundToWorkArea(
	SpriteWorldPtr spriteWorldP);

SW_FUNC void SWUpdateWindow(
	SpriteWorldPtr spriteWorldP);

SW_FUNC void SWUpdateSpriteWorld(
	SpriteWorldPtr spriteWorldP,
	Boolean updateWindow);

SW_FUNC void SWProcessSpriteWorld(
	SpriteWorldPtr spriteWorldP);

SW_FUNC void SWAnimateSpriteWorld(
	SpriteWorldPtr spriteWorldP);

SW_FUNC void  SWCheckIdleSpriteOverlap(
	SpriteWorldPtr			spriteWorldP,
	register SpritePtr		idleSpriteP,
	SpritePtr				headActiveSpriteP);

SW_FUNC void SWFindSpritesToBeRemoved(
	SpriteWorldPtr spriteWorldP );

SW_FUNC void SWSetSpriteWorldMaxFPS(
	SpriteWorldPtr spriteWorldP,
	short framesPerSec);

SW_FUNC OSErr SWSyncSpriteWorldToVBL(
	SpriteWorldPtr	spriteWorldP,
	Boolean			syncingOn);

SW_FUNC unsigned long SWGetSpriteWorldVersion( void );


#if GENERATINGCFM
void		SWVBLTask(VBLTaskRecPtr vblTaskPtr);/* Under CFM, we're passed the VBL task pointer */
#else
void		SWVBLTask(void);					/* Otherwise, we'll have to get it out of register A0 */ 
#endif


#ifdef __cplusplus
}
#endif

#endif /* __SPRITEWORLD__ */